﻿// <auto-generated />

using Microsoft.Xrm.Sdk;
using Microsoft.Xrm.Sdk.Client;
using System;
using System.Collections.ObjectModel;
using System.Linq;
using static System.FormattableString;

namespace NamespaceName
{
    /// <summary>
    /// Base class used for plugins
    /// </summary>
    [System.CodeDom.Compiler.GeneratedCodeAttribute("XrmToolkit", "XrmToolkitVersion")]
    public abstract partial class BasePluginClassName : IPlugin
    {
        protected enum Stage
        {
            PreValidation = 10,
            PreOperation = 20,
            MainOperation = 30,
            PostOperation = 40
        }

        /// <summary>
        /// Initializes a new instance of the BasePluginClassName class.
        /// </summary>
        internal BasePluginClassName(string unsecureConfig, string secureConfig)
        {
            this.UnsecureConfig = unsecureConfig;
            this.SecureConfig = secureConfig;
        }

        /// <summary>
        /// Un secure configuration specified during the registration of the plugin step
        /// </summary>
        public string UnsecureConfig { get; }

        /// <summary>
        /// Secure configuration specified during the registration of the plugin step
        /// </summary>
        public string SecureConfig { get; }

        /// <summary>
        /// Collection of events for which the plug-in should execute.
        /// </summary>
        protected Collection<PluginEvent> RegisteredEvents = new Collection<PluginEvent>();

        /// <summary>
        /// Executes the plug-in.
        /// </summary>
        /// <param name="serviceProvider">The service provider.</param>
        /// <remarks>
        /// For improved performance, Microsoft Dynamics CRM caches plug-in instances. 
        /// The plug-in's Execute method should be written to be stateless as the constructor 
        /// is not called for every invocation of the plug-in. Also, multiple system threads 
        /// could execute the plug-in at the same time. All per invocation state information 
        /// is stored in the context. This means that you should not use global variables in plug-ins.
        /// </remarks>
        public virtual void Execute(IServiceProvider serviceProvider)
        {
            if (serviceProvider == null)
                throw new InvalidPluginExecutionException($"The plugin is executing without an '{nameof(IServiceProvider)}'.", new ArgumentNullException(nameof(serviceProvider)));

            var tracingService = (ITracingService)serviceProvider.GetService(typeof(ITracingService));
            var pluginContext = (IPluginExecutionContext)serviceProvider.GetService(typeof(IPluginExecutionContext));

            tracingService.Trace(Invariant($"Entered {this.GetType()}.{nameof(Execute)}()"));
            try
            {
                // Iterate over all of the expected registered events to ensure that the plugin
                // has been invoked by an expected event
                var entityActions =
                    this.RegisteredEvents
                    .Where(a =>
                        (int)a.Stage == pluginContext.Stage &&
                        (string.IsNullOrWhiteSpace(a.MessageName) ? true : a.MessageName.ToLowerInvariant() == pluginContext.MessageName.ToLowerInvariant()) &&
                        (string.IsNullOrWhiteSpace(a.TableName) ? true : a.TableName.ToLowerInvariant() == pluginContext.PrimaryEntityName.ToLowerInvariant())
                    ).Select(a => a.PluginAction);

                if (entityActions.Any())
                {
                    foreach (var entityAction in entityActions)
                    {
                        tracingService.Trace(Invariant($"{this.GetType()} is firing for Entity: {pluginContext.PrimaryEntityName}, Message: {pluginContext.MessageName}, Method: {entityAction.Method.Name}"));
                        entityAction.Invoke(serviceProvider);
                    }
                }
            }
            catch (InvalidPluginExecutionException ex)
            {
                tracingService.Trace(Invariant($"Exception: {ex}"));
                throw;
            }
            catch (Exception ex)
            {
                tracingService.Trace(Invariant($"Exception: {ex}"));
                throw new InvalidPluginExecutionException(ex.Message, ex);
            }
            finally
            {
                tracingService.Trace(Invariant($"Exiting {this.GetType()}.{nameof(Execute)}()"));
            }
        }

        protected partial class PluginEvent
        {
            /// <summary>
            /// Execution pipeline stage that the plugin should be registered against.
            /// </summary>
            public Stage Stage { get; set; }
            /// <summary>
            /// Logical name of the entity that the plugin should be registered against. Leave 'null' to register against all entities.
            /// </summary>
            public string TableName { get; set; }
            /// <summary>
            /// Name of the message that the plugin should be triggered off of.
            /// </summary>
            public string MessageName { get; set; }
            /// <summary>
            /// Method that should be executed when the conditions of the Plugin Event have been met.
            /// </summary>
            public Action<IServiceProvider> PluginAction { get; set; }
        }

        protected partial class LocalPluginContext<T> : IDisposable
            where T : Entity
        {
            internal LocalPluginContext(IServiceProvider serviceProvider)
            {
                if (serviceProvider == null)
                    throw new ArgumentNullException(nameof(serviceProvider));

                // Obtain the tracing service from the service provider.
                this.TracingService = (ITracingService)serviceProvider.GetService(typeof(ITracingService));

                // Obtain the execution context service from the service provider.
                this.PluginExecutionContext = (IPluginExecutionContext)serviceProvider.GetService(typeof(IPluginExecutionContext));

                // Obtain the Organization Service factory service from the service provider
                this.ServiceFactory = (IOrganizationServiceFactory)serviceProvider.GetService(typeof(IOrganizationServiceFactory));

                // Use the factory to generate the Organization Service.
                this.OrganizationService = this.ServiceFactory.CreateOrganizationService(this.PluginExecutionContext.UserId);

                // Generate the CrmContext to use with LINQ etc
                this.CrmContext = new OrganizationServiceContext(this.OrganizationService);
            }

            internal OrganizationServiceContext CrmContext { get; private set; }

            internal IServiceProvider ServiceProvider { get; }

            internal IOrganizationServiceFactory ServiceFactory { get; }

            internal IOrganizationService OrganizationService { get; }

            internal IPluginExecutionContext PluginExecutionContext { get; }

            internal ITracingService TracingService { get; }

            internal Stage Stage => (Stage)this.PluginExecutionContext.Stage;

            internal int Depth => this.PluginExecutionContext.Depth;

            internal string MessageName => this.PluginExecutionContext.MessageName;

            internal void Trace(string message)
            {
                if (string.IsNullOrWhiteSpace(message) || this.TracingService == null)
                    return;

                if (this.PluginExecutionContext == null)
                    this.TracingService.Trace(message);
                else
                    this.TracingService.Trace(Invariant($"{message}, Correlation Id: {this.PluginExecutionContext.CorrelationId}, Initiating User: {this.PluginExecutionContext.InitiatingUserId}"));
            }

            public void Dispose()
            {
                if (this.CrmContext != null)
                {
                    this.CrmContext.Dispose();
                    this.CrmContext = null;
                }
            }

            /// <summary>
            /// Returns the first registered 'Pre' image for the pipeline execution
            /// </summary>
            internal virtual T PreImage
            {
                get
                {
                    if (this.PluginExecutionContext.PreEntityImages.Any())
                        return GetEntityAsType(this.PluginExecutionContext.PreEntityImages[this.PluginExecutionContext.PreEntityImages.FirstOrDefault().Key]);
                    return null;
                }
            }

            /// <summary>
            /// Returns the first registered 'Post' image for the pipeline execution
            /// </summary>
            internal virtual T PostImage
            {
                get
                {
                    if (this.PluginExecutionContext.PostEntityImages.Any())
                        return GetEntityAsType(this.PluginExecutionContext.PostEntityImages[this.PluginExecutionContext.PostEntityImages.FirstOrDefault().Key]);
                    return null;
                }
            }

            /// <summary>
            /// Returns the 'Target' of the message if available
            /// This is an 'Entity' instead of the specified type in order to retain the same instance of the 'Entity' object. This allows for updates to the target in a 'Pre' stage that
            /// will get persisted during the transaction.
            /// </summary>
            internal virtual Entity TargetEntity
            {
                get
                {
                    if (this.PluginExecutionContext.InputParameters.Contains("Target"))
                        return this.PluginExecutionContext.InputParameters["Target"] as Entity;
                    return null;
                }
            }

            /// <summary>
            /// Returns the 'Target' of the message as an EntityReference if available
            /// </summary>
            internal virtual EntityReference TargetEntityReference
            {
                get
                {
                    if (this.PluginExecutionContext.InputParameters.Contains("Target"))
                        return this.PluginExecutionContext.InputParameters["Target"] as EntityReference;
                    return null;
                }
            }

            private T GetEntityAsType(Entity entity)
            {
                if (typeof(T) == entity.GetType())
                    return entity as T;
                else
                    return entity.ToEntity<T>();
            }
        }
    }
}
