﻿' <auto-generated />

Imports System.Collections.ObjectModel
Imports Microsoft.Xrm.Sdk
Imports Microsoft.Xrm.Sdk.Client
Imports System.FormattableString

Namespace NamespaceName
    ''' <summary>
    ''' Base class used for plugins
    ''' </summary>
    <System.CodeDom.Compiler.GeneratedCodeAttribute("XrmToolkit", "XrmToolkitVersion")>
    Partial Public MustInherit Class BasePluginClassName
        Implements IPlugin

        Protected Enum Stage
            PreValidation = 10
            PreOperation = 20
            MainOperation = 30
            PostOperation = 40
        End Enum

        ''' <summary>
        ''' Initializes a new instance of the BasePluginClassName class.
        ''' </summary>
        Friend Sub New(unsecureConfig As String, secureConfig As String)
            Me.UnsecureConfig = unsecureConfig
            Me.SecureConfig = secureConfig
        End Sub

        ''' <summary>
        ''' Un secure configuration specified during the registration of the plugin step
        ''' </summary>
        Public ReadOnly Property UnsecureConfig As String

        ''' <summary>
        ''' Secure configuration specified during the registration of the plugin step
        ''' </summary>
        Public ReadOnly Property SecureConfig As String

        ''' <summary>
        ''' Collection of events for which the plug-in should execute.
        ''' </summary>
        Protected ReadOnly Property RegisteredEvents As Collection(Of PluginEvent) = New Collection(Of PluginEvent)()

        ''' <summary>
        ''' Executes the plug-in.
        ''' </summary>
        ''' <param name="serviceProvider">The service provider.</param>
        ''' <remarks>
        ''' For improved performance, Microsoft Dynamics CRM caches plug-in instances. 
        ''' The plug-in's Execute method should be written to be stateless as the constructor 
        ''' is not called for every invocation of the plug-in. Also, multiple system threads 
        ''' could execute the plug-in at the same time. All per invocation state information 
        ''' is stored in the context. This means that you should not use global variables in plug-ins.
        ''' </remarks>
        Public Overridable Sub Execute(serviceProvider As IServiceProvider) Implements IPlugin.Execute
            If serviceProvider Is Nothing Then
                Throw New InvalidPluginExecutionException($"The plugin is executing without an '{NameOf(IServiceProvider)}'.", New ArgumentNullException(NameOf(serviceProvider)))
            End If

            Dim tracingService = DirectCast(serviceProvider.GetService(GetType(ITracingService)), ITracingService)
            Dim pluginContext = DirectCast(serviceProvider.GetService(GetType(IPluginExecutionContext)), IPluginExecutionContext)

            tracingService.Trace(Invariant($"Entered {Me.GetType()}.{NameOf(Execute)}()"))
            Try
                ' Iterate over all of the expected registered events to ensure that the plugin
                ' has been invoked by an expected event
                Dim entityActions = (From a In Me.RegisteredEvents Where (CInt(a.Stage) = pluginContext.Stage AndAlso (If(String.IsNullOrWhiteSpace(a.MessageName), True, a.MessageName.ToLowerInvariant() = pluginContext.MessageName.ToLowerInvariant())) AndAlso (If(String.IsNullOrWhiteSpace(a.TableName), True, a.TableName.ToLowerInvariant() = pluginContext.PrimaryEntityName.ToLowerInvariant()))) Select a.PluginAction)

                If entityActions.Any() Then
                    For Each entityAction As Action(Of IServiceProvider) In entityActions
                        tracingService.Trace(Invariant($"{Me.GetType()} is firing for Entity: {pluginContext.PrimaryEntityName}, Message: {pluginContext.MessageName}, Method: {entityAction.Method.Name}"))
                        entityAction.Invoke(serviceProvider)
                    Next
                End If
            Catch ex As InvalidPluginExecutionException
                tracingService.Trace(Invariant($"Exception: {ex}"))
                Throw
            Catch ex As Exception
                tracingService.Trace(Invariant($"Exception: {ex}"))
                Throw New InvalidPluginExecutionException(ex.Message, ex)
            Finally
                tracingService.Trace(Invariant($"Exiting {Me.GetType()}.{NameOf(Execute)}()"))
            End Try
        End Sub

        Partial Protected Class PluginEvent
            ''' <summary>
            ''' Execution pipeline stage that the plugin should be registered against.
            ''' </summary>
            Public Property Stage As Stage

            ''' <summary>
            ''' Logical name of the entity that the plugin should be registered against. Leave 'null' to register against all entities.
            ''' </summary>
            Public Property TableName As String

            ''' <summary>
            ''' Name of the message that the plugin should be triggered off of.
            ''' </summary>
            Public Property MessageName As String

            ''' <summary>
            ''' Method that should be executed when the conditions of the Plugin Event have been met.
            ''' </summary>
            Public Property PluginAction As Action(Of IServiceProvider)
        End Class

        Partial Protected Class LocalPluginContext(Of T As Entity)
            Implements IDisposable

            Friend Sub New(serviceProvider As IServiceProvider)
                If serviceProvider Is Nothing Then
                    Throw New ArgumentNullException(NameOf(serviceProvider))
                End If

                ' Obtain the tracing service from the service provider.
                Me.TracingService = DirectCast(serviceProvider.GetService(GetType(ITracingService)), ITracingService)

                ' Obtain the execution context service from the service provider.
                Me.PluginExecutionContext = DirectCast(serviceProvider.GetService(GetType(IPluginExecutionContext)), IPluginExecutionContext)

                ' Obtain the Organization Service factory service from the service provider
                Me.ServiceFactory = DirectCast(serviceProvider.GetService(GetType(IOrganizationServiceFactory)), IOrganizationServiceFactory)

                ' Use the factory to generate the Organization Service.
                Me.OrganizationService = Me.ServiceFactory.CreateOrganizationService(Me.PluginExecutionContext.UserId)

                ' Generate the CrmContext to use with LINQ etc
                Me.CrmContext = New OrganizationServiceContext(Me.OrganizationService)
            End Sub

            Friend Property CrmContext As OrganizationServiceContext

            Friend ReadOnly Property ServiceProvider As IServiceProvider

            Friend ReadOnly Property ServiceFactory As IOrganizationServiceFactory

            Friend ReadOnly Property OrganizationService As IOrganizationService

            Friend ReadOnly Property PluginExecutionContext As IPluginExecutionContext

            Friend ReadOnly Property TracingService As ITracingService

            Friend ReadOnly Property Stage As Stage
                Get
                    Return CType(Me.PluginExecutionContext.Stage, Stage)
                End Get
            End Property

            Friend ReadOnly Property Depth As Integer
                Get
                    Return Me.PluginExecutionContext.Depth
                End Get
            End Property

            Friend ReadOnly Property MessageName As String
                Get
                    Return Me.PluginExecutionContext.MessageName
                End Get
            End Property

            Friend Sub Trace(message As String)
                If String.IsNullOrWhiteSpace(message) OrElse Me.TracingService Is Nothing Then
                    Return
                End If

                If Me.PluginExecutionContext Is Nothing Then
                    Me.TracingService.Trace(message)
                Else
                    Me.TracingService.Trace(Invariant($"{message}, Correlation Id: {PluginExecutionContext.CorrelationId}, Initiating User: {PluginExecutionContext.InitiatingUserId}"))
                End If
            End Sub

            Public Sub Dispose() Implements IDisposable.Dispose
                If Me.CrmContext IsNot Nothing Then
                    Me.CrmContext.Dispose()
                    Me.CrmContext = Nothing
                End If
            End Sub

            ''' <summary>
            ''' Returns the first registered 'Pre' image for the pipeline execution
            ''' </summary>
            Friend Overridable ReadOnly Property PreImage As T
                Get
                    If Me.PluginExecutionContext.PreEntityImages.Any() Then
                        Return GetEntityAsType(Me.PluginExecutionContext.PreEntityImages(Me.PluginExecutionContext.PreEntityImages.FirstOrDefault().Key))
                    End If
                    Return Nothing
                End Get
            End Property

            ''' <summary>
            ''' Returns the first registered 'Post' image for the pipeline execution
            ''' </summary>
            Friend Overridable ReadOnly Property PostImage As T
                Get
                    If Me.PluginExecutionContext.PostEntityImages.Any() Then
                        Return GetEntityAsType(Me.PluginExecutionContext.PostEntityImages(Me.PluginExecutionContext.PostEntityImages.FirstOrDefault().Key))
                    End If
                    Return Nothing
                End Get
            End Property

            ''' <summary>
            ''' Returns the 'Target' of the message if available
            ''' This is an 'Entity' instead of the specified type in order to retain the same instance of the 'Entity' object. This allows for updates to the target in a 'Pre' stage that
            ''' will get persisted during the transaction.
            ''' </summary>
            Friend Overridable ReadOnly Property TargetEntity As Entity
                Get
                    If Me.PluginExecutionContext.InputParameters.Contains("Target") Then
                        Return TryCast(Me.PluginExecutionContext.InputParameters("Target"), Entity)
                    End If
                    Return Nothing
                End Get
            End Property

            ''' <summary>
            ''' Returns the 'Target' of the message as an EntityReference if available
            ''' </summary>
            Friend Overridable ReadOnly Property TargetEntityReference As EntityReference
                Get
                    If Me.PluginExecutionContext.InputParameters.Contains("Target") Then
                        Return TryCast(Me.PluginExecutionContext.InputParameters("Target"), EntityReference)
                    End If
                    Return Nothing
                End Get
            End Property

            Private Function GetEntityAsType(entity As Entity) As T
                If GetType(T) = entity.GetType() Then
                    Return TryCast(entity, T)
                Else
                    Return entity.ToEntity(Of T)()
                End If
            End Function
        End Class
    End Class
End Namespace
