﻿using Microsoft.Crm.Sdk.Messages;
using Microsoft.Xrm.Sdk;
using Microsoft.Xrm.Sdk.Messages;
using Microsoft.Xrm.Sdk.Metadata;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
namespace ProxyClassNameSpace
{
    class EntityMethods : _BaseProxyClass_
    {
        public EntityMethods() : base(new Entity()) { }

        /// <summary>
        /// Adds an item to a campaign.
        /// </summary>
        /// <param name="Service">Organization Service</param>
        /// <param name="Item">Item to add to the Campaign</param>
        /// <returns></returns>
        public Guid AddItemCampaignRequest(IOrganizationService Service, EntityReference Item)
        {
            AddItemCampaignRequest Request = new AddItemCampaignRequest() { CampaignId = this.Id, EntityId = Item.Id, EntityName = Item.LogicalName };
            return ((AddItemCampaignResponse)Service.Execute(Request)).CampaignItemId;
        }
        /// <summary>
        /// Adds an item to a campaign activity.
        /// </summary>
        /// <param name="Service">Organization Service</param>
        /// <param name="Item">Item to add to the Campaign Activity</param>
        /// <returns></returns>
        public Guid AddItemCampaignActivityRequest(IOrganizationService Service, EntityReference Item)
        {
            AddItemCampaignActivityRequest Request = new AddItemCampaignActivityRequest() { CampaignActivityId = this.Id, ItemId = Item.Id, EntityName = Item.LogicalName };
            return ((AddItemCampaignActivityResponse)Service.Execute(Request)).CampaignActivityItemId;
        }
        /// <summary>
        /// Adds a group of members to a list.
        /// </summary>
        /// <param name="Service">Organization Service</param>
        /// <param name="EntityIds">Ids of the members to add to the list</param>
        public void AddListMembersListRequest(IOrganizationService Service, List<Guid> MemberIds)
        {
            AddListMembersListRequest Request = new AddListMembersListRequest() { ListId = this.Id, MemberIds = MemberIds.ToArray() };
            Service.Execute(Request);
        }
        /// <summary>
        /// Adds a member to a list.
        /// </summary>
        /// <param name="Service">Organization Service</param>
        /// <param name="Entity">Id of the member to add to the list</param>
        public void AddMemberListRequest(IOrganizationService Service, Guid Entity)
        {
            AddMemberListRequest Request = new AddMemberListRequest() { ListId = this.Id, EntityId = Entity };
            Service.Execute(Request);
        }
        /// <summary>
        /// Add a group of members to a team.
        /// </summary>
        /// <param name="Service">Organization Service</param>
        /// <param name="MemberIds">Ids of the members to add to the team.</param>
        public void AddMembersTeamRequest(IOrganizationService Service, List<Guid> MemberIds)
        {
            AddMembersTeamRequest Request = new AddMembersTeamRequest() { TeamId = this.Id, MemberIds = MemberIds.ToArray() };
            Service.Execute(Request);
        }
        /// <summary>
        /// Add privileges to the current role.
        /// </summary>
        /// <param name="Service">Organization Service</param>
        /// <param name="Privileges">Privileges to add to the current role.</param>
        public void AddPrivilegesRoleRequest1(IOrganizationService Service, List<RolePrivilege> Privileges)
        {
            AddPrivilegesRoleRequest Request = new AddPrivilegesRoleRequest() { RoleId = this.Id, Privileges = Privileges.ToArray() };
            Service.Execute(Request);
        }
        /// <summary>
        /// Add a privilege to the current role.
        /// </summary>
        /// <param name="Service">Organization Service</param>
        /// <param name="EntityLogicalName">Entity Logical Name</param>
        /// <param name="Type">Privilege Type</param>
        /// <param name="Depth">Privilege Level</param>
        public void AddPrivilegesRoleRequest2(IOrganizationService Service, string EntityLogicalName, PrivilegeType Type, PrivilegeDepth Depth)
        {
            RetrieveEntityRequest RERequest = new RetrieveEntityRequest() { LogicalName = base.LogicalName, EntityFilters = Microsoft.Xrm.Sdk.Metadata.EntityFilters.Privileges };
            RetrieveEntityResponse REResponse = Service.Execute(RERequest) as RetrieveEntityResponse;

            SecurityPrivilegeMetadata Privilege = (from Priv in REResponse.EntityMetadata.Privileges where Priv.PrivilegeType == Type select Priv).FirstOrDefault();
            if (Privilege != null)
            {
                AddPrivilegesRoleRequest Request = new AddPrivilegesRoleRequest()
                {
                    RoleId = this.Id,
                    Privileges = new RolePrivilege[] {
                        new RolePrivilege{ BusinessUnitId = this.GetAttributeValue<Guid>("businessunitid"), PrivilegeId = Privilege.PrivilegeId, Depth = Depth} 
                    }
                };
                Service.Execute(Request);
            }
        }
        /// <summary>
        /// Remove a privilege from the current role.
        /// </summary>
        /// <param name="Service">Organization Service</param>
        /// <param name="Privilege">Id of the privilege to remove.</param>
        public void RemovePrivilegeRoleRequest(IOrganizationService Service, Guid Privilege)
        {
            RemovePrivilegeRoleRequest Request = new RemovePrivilegeRoleRequest() { RoleId = this.Id, PrivilegeId = Privilege };
            Service.Execute(Request);
        }
        /// <summary>
        /// Add the current item to a specified queue.
        /// </summary>
        /// <param name="Service">Organization Service</param>
        /// <param name="DestinationQueueId">Queue to add/move the item to</param>
        /// <param name="SourceQueueId">Queue to move the item from (Optional)</param>
        /// <returns>Id of the Queue Item</returns>
        public Guid AddToQueueRequest(IOrganizationService Service, Guid DestinationQueueId, Guid SourceQueueId = default(Guid))
        {
            AddToQueueRequest Request = new AddToQueueRequest() { DestinationQueueId = DestinationQueueId, SourceQueueId = SourceQueueId, Target = this };
            return ((AddToQueueResponse)Service.Execute(Request)).QueueItemId;
        }
        /// <summary>
        /// Add an item to the current queue.
        /// </summary>
        /// <param name="Service">Organization Service</param>
        /// <param name="Item">Item to be added/moved to the current queue</param>
        /// <param name="SourceQueueId">Queue to move the item from (Optional)</param>
        /// <returns>Id of the Queue Item</returns>
        public Guid AddItemToQueueRequest(IOrganizationService Service, EntityReference Item, Guid SourceQueueId = default(Guid))
        {
            AddToQueueRequest Request = new AddToQueueRequest() { DestinationQueueId = this.Id, SourceQueueId = SourceQueueId, Target = Item };
            return ((AddToQueueResponse)Service.Execute(Request)).QueueItemId;
        }
        protected override _BaseProxyClass_.eTextOptions GetTextOptions(string LogicalName)
        {
            throw new NotImplementedException();
        }
        protected override string GetErrorString(string AttributeName, _BaseProxyClass_.eErrorType DefaultErrorType)
        {
            throw new NotImplementedException();
        }
        protected override _BaseProxyClass_.eNumberOptions GetNumberOptions(string LogicalName)
        {
            throw new NotImplementedException();
        }
    }
}
