﻿declare namespace XrmBase {
    /**
     * Lookup-like type object for userSettings.roles.
     */
    interface Role {
        id: string;
        name: string;
    }

    interface userSettings {
        /**
         * Collection of lookup-like objects containing the GUID and display name of each of the security role or teams that the user is associated with.
         */
        roles: Collection<Role>;

        /**
         * Returns an array of strings that represent the GUID values of each of the security role privilege that the user is associated with or any teams that the user is associated with.
         */
        securityRolePrivileges: string[]
    }

    /**
     * Interface for the ui of a form.
     */
    interface UiModule<T extends TabCollectionBase, U extends ControlCollectionBase> {
        /**
         * Section containing the header items.
         */
        headerSection: HeaderSection;

        /**
         * Section containing the footer items.
         */
        footerSection: FooterSection;
    }

    interface HeaderSection {
        /**
         * Returns the header's body visibility.
         */
        getBodyVisible(): boolean;

        /**
         * Returns the command bar visibility.
         */
        getCommandBarVisible(): boolean;

        /**
         * Returns the tab navigator visibility.
         */
        getTabNavigatorVisible(): boolean;

        /**
         * Sets the header's body visibility.
         */
        setBodyVisible(isVisible: boolean);

        /**
         * Sets the command bar visibility.
         */
        setCommandBarVisible(isVisible: boolean);

        /**
         * Sets the tab navigator visibility.
         */
        setTabNavigatorVisible(isVisible: boolean);
    }

    interface FooterSection {
        /**
         * Gets the visibility of the footer.
         */
        getVisible(): boolean;

        /**
         * Sets the visibility of the footer.
         * @param isVisible
         */
        setVisible(isVisible: boolean);
    }

    type PageTabContentType = "cardSections" | "singleComponent";

    interface PageTab<T extends SectionCollectionBase> {
        /**
         * Returns the content type.
         */
        getContentType(): PageTabContentType;

        /**
         * Sets the content type.
         * @param contentType
         */
        setContentType(contentType: PageTabContentType): void;
    }
}
