﻿' <auto-generated />

Imports System.Linq
Imports System.Text
Imports System.Collections
Imports System.Reflection
Imports System.Collections.Generic
Imports Microsoft.Xrm.Sdk
Imports Microsoft.Xrm.Sdk.Query
Imports Microsoft.Crm.Sdk.Messages
Imports Microsoft.Xrm.Sdk.Messages
Imports Microsoft.Xrm.Sdk.Metadata
Imports System.ComponentModel
Imports System.Runtime.Serialization

<Assembly: Microsoft.Xrm.Sdk.Client.ProxyTypesAssemblyAttribute()>
Namespace ProxyClassNameSpace
    ''' <summary>
    ''' Base class used for proxy classes
    ''' </summary>
    <DataContract(Name:="Entity", Namespace:="http://schemas.microsoft.com/xrm/2011/Contracts")>
    Partial Public MustInherit Class _BaseProxyClass_
        Inherits Entity
        Implements INotifyPropertyChanged
        Implements INotifyPropertyChanging
        Protected Sub New(original As Entity)
            If String.IsNullOrEmpty(original.LogicalName) Then
                Throw New Exception("Please specify the 'LogicalName' on the entity when using a proxy class.")
            End If
            Dim EntityName As String = GetLogicalName(Me.[GetType]())
            If EntityName <> original.LogicalName Then
                Throw New Exception("Please make sure that the entity logical name matches that of the proxy class you are creating.")
            End If

            Me.LogicalName = original.LogicalName
            Me.RelatedEntities.Clear()
            Me.FormattedValues.Clear()
            Me.Attributes = original.Attributes
            Me.RelatedEntities.AddRange(original.RelatedEntities)
            Me.FormattedValues.AddRange(original.FormattedValues)
            Me.ExtensionData = original.ExtensionData
            Me.EntityState = original.EntityState
            Me.RowVersion = original.RowVersion
            Me.KeyAttributes = original.KeyAttributes
            Me.Id = original.Id
        End Sub
        Public Shared Property ReturnDatesInLocalTime As Boolean = False

        ''' <summary>
        ''' Indicates if the entity was retrieved from CRM and whether all columns were retrieved.
        ''' </summary>
        ''' <returns></returns>
        Public Property AllColumnsRetrievedFromCrm As Boolean = False

        Public Function Create(service As IOrganizationService) As Guid
            Me.Id = service.Create(Me)
            ChangedValues.Clear()
            Return Me.Id
        End Function
        Public Sub Delete(service As IOrganizationService)
            service.Delete(Me.LogicalName, Me.Id)
        End Sub
        Public Sub Update(service As IOrganizationService)
            If ChangedValues.Count > 0 Then
                service.Update(GetChangedEntity())
            End If
            ChangedValues.Clear()
        End Sub
        Public Async Function CreateAsync(service As IOrganizationService) As System.Threading.Tasks.Task(Of Guid)
            Return Await System.Threading.Tasks.Task.Run(Function()
                                                             Return Me.Create(service)
                                                         End Function)
        End Function
        Public Async Function UpdateAsync(service As IOrganizationService) As System.Threading.Tasks.Task
            Await System.Threading.Tasks.Task.Run(Sub()
                                                      Me.Update(service)
                                                  End Sub)
        End Function
        Public Async Function DeleteAsync(service As IOrganizationService) As System.Threading.Tasks.Task
            Await System.Threading.Tasks.Task.Run(Sub()
                                                      Me.Delete(service)
                                                  End Sub)
        End Function
        Public Function GetChangedEntity() As Entity
            Dim entity As New Entity(Me.LogicalName)
            entity.Id = Me.Id
            For Each attributeName As String In ChangedValues.Keys
                entity.Attributes(attributeName) = Me.Attributes(attributeName)
            Next
            Return entity
        End Function
        Public Sub Save(service As IOrganizationService)
            If Me.Id <> Guid.Empty Then
                Me.Update(service)
            Else
                Me.Create(service)
            End If
        End Sub
        Public Async Function SaveAsync(service As IOrganizationService) As System.Threading.Tasks.Task
            Await System.Threading.Tasks.Task.Run(Sub()
                                                      Me.Save(service)
                                                  End Sub)
        End Function
        Public Shared Function CreateProxyType(entity As Entity) As _BaseProxyClass_
            If Not _proxyTypes.ContainsValue(entity.LogicalName) Then
                'Register all the types in the current assembly
                RegisterProxyTypesInAssembly(GetType(_BaseProxyClass_).Assembly)
            End If
            If _proxyTypes.ContainsValue(entity.LogicalName) Then
                Dim proxyType As Type = (From p In _proxyTypes.Keys Where _proxyTypes(p) = entity.LogicalName Select p).FirstOrDefault()
                Dim constructor As ConstructorInfo = proxyType.GetConstructor(New Type() {GetType(Entity)})
                Return DirectCast(constructor.Invoke(New Object() {entity}), _BaseProxyClass_)
            Else
                Throw New Exception("The proxy class for type '" + entity.LogicalName + "' has not been registered.")
            End If
        End Function
        Private Shared _proxyTypes As New Dictionary(Of Type, String)()
        Public Shared Sub RegisterProxyType(proxyType As Type, logicalName As String)
            If Not _proxyTypes.ContainsKey(proxyType) Then
                _proxyTypes.Add(proxyType, logicalName)
            End If
        End Sub
        Public Shared Sub RegisterProxyTypesInAssembly(assembly As Assembly)
            If (assembly IsNot Nothing) Then
                For Each type As Type In assembly.GetTypes()
                    Runtime.CompilerServices.RuntimeHelpers.RunClassConstructor(type.TypeHandle)
                Next
            End If
        End Sub
        Public Shared Function GetLogicalName(Of T As _BaseProxyClass_)() As String
            Return GetLogicalName(GetType(T))
        End Function
        Public Shared Function GetLogicalName(type As Type) As String
            If Not _proxyTypes.ContainsKey(type) Then
                'This forces the static constructor to be called and register the name
                Dim constructor As ConstructorInfo = type.GetConstructor(New Type() {})
                constructor.Invoke(New Object() {})
            End If
            Return _proxyTypes(type)
        End Function
        Public Shared Widening Operator CType(proxy As _BaseProxyClass_) As EntityReference
            If proxy IsNot Nothing Then
                Return proxy.ToEntityReference()
            End If
            Return Nothing
        End Operator
        Public Function GetPropertyValue(Of T)(name As String) As T
            If Me.Contains(name) Then
                Dim value As T = DirectCast(Me.Attributes(name), T)
                If (GetType(T) = GetType(DateTime) OrElse GetType(T) = GetType(System.Nullable(Of DateTime))) AndAlso ReturnDatesInLocalTime AndAlso value IsNot Nothing Then
                    value = DirectCast(DirectCast(DirectCast(DirectCast(value, [Object]), DateTime).ToLocalTime(), [Object]), T)
                End If
                Return value
            End If
            Return Nothing
        End Function
        ''' <summary>
        ''' Collection of values that have changed along with their original value.
        ''' </summary>
        ''' <returns></returns>
        Protected ReadOnly Property ChangedValues As New Dictionary(Of String, Object)()
        Private ReadOnly _equalityComparer As New AttributeEqualityComparer()
        Public Overridable Sub SetPropertyValue(Of T)(name As String, value As T, propertyName As String)
            Dim currentValue As Object = If(Me.Contains(name), DirectCast(Me.GetPropertyValue(Of T)(name), Object), Nothing)
            If (currentValue Is Nothing AndAlso Me.AllColumnsRetrievedFromCrm = False) OrElse Not _equalityComparer.Equals(currentValue, value) Then
                OnPropertyChanging(propertyName)
                Dim originalValue As Object = Nothing
                If ChangedValues.TryGetValue(name, originalValue) Then
                    If _equalityComparer.Equals(originalValue, value) Then
                        ChangedValues.Remove(name)
                    End If
                Else
                    ChangedValues.Add(name, currentValue)
                End If
                Me.Attributes(name) = value
                OnPropertyChanged(propertyName)
            End If
        End Sub
        Protected Sub OnPropertyChanging(propertyName As String)
            RaiseEvent PropertyChanging(Me, New PropertyChangingEventArgs(propertyName))
        End Sub
        Protected Sub OnPropertyChanged(propertyName As String)
            RaiseEvent PropertyChanged(Me, New PropertyChangedEventArgs(propertyName))
        End Sub
        Public Overridable Sub SetPropertyValue(name As String, value As String, maxLength As Integer, propertyName As String)
            Dim textOptions As eTextOptions = GetTextOptions(name)
            If textOptions <> eTextOptions.Ignore AndAlso Not String.IsNullOrEmpty(value) AndAlso value.Length > maxLength Then
                If textOptions = eTextOptions.Truncate Then
                    value = value.Substring(0, maxLength)
                Else
                    Throw New Exception(String.Format(GetErrorString(name, eErrorType.Text), name, value, value.Length, maxLength))
                End If
            End If
            SetPropertyValue(Of String)(name, value, propertyName)
        End Sub
        Public Overridable Sub SetPropertyValue(name As String, value As System.Nullable(Of Integer), minValue As Integer, maxValue As Integer, propertyName As String)
            Dim numberOptions As eNumberOptions = GetNumberOptions(name)
            If numberOptions <> eNumberOptions.Ignore AndAlso (value < minValue OrElse value > maxValue) Then
                Dim throwError As Boolean = False
                If numberOptions = eNumberOptions.CorrectMinAndMax Then
                    value = If((value < minValue), minValue, maxValue)
                ElseIf numberOptions = eNumberOptions.CorrectMinIgnoreMax Then
                    value = If((value < minValue), minValue, value)
                ElseIf numberOptions = eNumberOptions.CorrectMinThrowMax AndAlso value < minValue Then
                    value = minValue
                ElseIf numberOptions = eNumberOptions.CorrectMaxIgnoreMin Then
                    value = If((value > maxValue), maxValue, value)
                ElseIf numberOptions = eNumberOptions.CorrectMaxThrowMin AndAlso value > maxValue Then
                    value = maxValue
                Else
                    throwError = True
                End If
                If throwError Then
                    Throw New Exception(String.Format(GetErrorString(name, eErrorType.Number), name, value, minValue, maxValue))
                End If
            End If
            SetPropertyValue(Of System.Nullable(Of Integer))(name, value, propertyName)
        End Sub
        Public Overridable Sub SetPropertyValue(name As String, value As System.Nullable(Of Decimal), minValue As Decimal, maxValue As Decimal, propertyName As String)
            Dim numberOptions As eNumberOptions = GetNumberOptions(name)
            If numberOptions <> eNumberOptions.Ignore AndAlso (value < minValue OrElse value > maxValue) Then
                Dim throwError As Boolean = False
                If numberOptions = eNumberOptions.CorrectMinAndMax Then
                    value = If((value < minValue), minValue, maxValue)
                ElseIf numberOptions = eNumberOptions.CorrectMinIgnoreMax Then
                    value = If((value < minValue), minValue, value)
                ElseIf numberOptions = eNumberOptions.CorrectMinThrowMax AndAlso value < minValue Then
                    value = minValue
                ElseIf numberOptions = eNumberOptions.CorrectMaxIgnoreMin Then
                    value = If((value > maxValue), maxValue, value)
                ElseIf numberOptions = eNumberOptions.CorrectMaxThrowMin AndAlso value > maxValue Then
                    value = maxValue
                Else
                    throwError = True
                End If
                If throwError Then
                    Throw New Exception(String.Format(GetErrorString(name, eErrorType.Number), name, value, minValue, maxValue))
                End If
            End If
            SetPropertyValue(Of System.Nullable(Of Decimal))(name, value, propertyName)
        End Sub
        Public Overridable Sub SetPropertyValue(name As String, value As System.Nullable(Of Double), minValue As Double, maxValue As Double, propertyName As String)
            Dim numberOptions As eNumberOptions = GetNumberOptions(name)
            If numberOptions <> eNumberOptions.Ignore AndAlso (value < minValue OrElse value > maxValue) Then
                Dim throwError As Boolean = False
                If numberOptions = eNumberOptions.CorrectMinAndMax Then
                    value = If((value < minValue), minValue, maxValue)
                ElseIf numberOptions = eNumberOptions.CorrectMinIgnoreMax Then
                    value = If((value < minValue), minValue, value)
                ElseIf numberOptions = eNumberOptions.CorrectMinThrowMax AndAlso value < minValue Then
                    value = minValue
                ElseIf numberOptions = eNumberOptions.CorrectMaxIgnoreMin Then
                    value = If((value > maxValue), maxValue, value)
                ElseIf numberOptions = eNumberOptions.CorrectMaxThrowMin AndAlso value > maxValue Then
                    value = maxValue
                Else
                    throwError = True
                End If
                If throwError Then
                    Throw New Exception(String.Format(GetErrorString(name, eErrorType.Number), name, value, minValue, maxValue))
                End If
            End If
            SetPropertyValue(Of System.Nullable(Of Double))(name, value, propertyName)
        End Sub
        Public Overridable Sub SetPropertyValue(name As String, value As Money, minValue As Decimal, maxValue As Decimal, propertyName As String)
            Dim numberOptions As eNumberOptions = GetNumberOptions(name)
            If value IsNot Nothing AndAlso numberOptions <> eNumberOptions.Ignore AndAlso (value.Value < minValue OrElse value.Value > maxValue) Then
                Dim throwError As Boolean = False
                If numberOptions = eNumberOptions.CorrectMinAndMax Then
                    value.Value = If((value.Value < minValue), minValue, maxValue)
                ElseIf numberOptions = eNumberOptions.CorrectMinIgnoreMax Then
                    value.Value = If((value.Value < minValue), minValue, value.Value)
                ElseIf numberOptions = eNumberOptions.CorrectMinThrowMax AndAlso value.Value < minValue Then
                    value.Value = minValue
                ElseIf numberOptions = eNumberOptions.CorrectMaxIgnoreMin Then
                    value.Value = If((value.Value > maxValue), maxValue, value.Value)
                ElseIf numberOptions = eNumberOptions.CorrectMaxThrowMin AndAlso value.Value > maxValue Then
                    value.Value = maxValue
                Else
                    throwError = True
                End If
                If throwError Then
                    Throw New Exception(String.Format(GetErrorString(name, eErrorType.Number), name, value.Value, minValue, maxValue))
                End If
            End If
            SetPropertyValue(Of Money)(name, value, propertyName)
        End Sub
        Protected MustOverride Function GetTextOptions(logicalName As String) As eTextOptions
        Protected MustOverride Function GetErrorString(attributeName As String, defaultErrorType As eErrorType) As String
        Protected MustOverride Function GetNumberOptions(logicalName As String) As eNumberOptions
        Public ReadOnly Property IsDirty() As Boolean
            Get
                Return Me.ChangedValues.Count > 0
            End Get
        End Property
        Public Shared Function GetRelatedOneToManyEntities(Of T As _BaseProxyClass_)(service As IOrganizationService, relatedId As Guid, logicalName As String, attributeName As String, ParamArray columns As String()) As List(Of T)
            Return GetRelatedOneToManyEntities(service, relatedId, logicalName, attributeName, columns).ToProxies(Of T)()
        End Function
        Public Shared Function GetRelatedOneToManyEntities(Of T As _BaseProxyClass_)(service As IOrganizationService, relatedId As Guid, logicalName As String, attributeName As String, columns As ColumnSet) As List(Of T)
            Return GetRelatedOneToManyEntities(service, relatedId, logicalName, attributeName, columns).ToProxies(Of T)()
        End Function
        Public Shared Function GetRelatedOneToManyEntities(service As IOrganizationService, relatedId As Guid, logicalName As String, attributeName As String, ParamArray columns As String()) As List(Of Entity)
            Dim ColumnSet As ColumnSet = If((columns.Length = 0), New ColumnSet(True), New ColumnSet(columns))
            Return GetRelatedOneToManyEntities(service, relatedId, logicalName, attributeName, ColumnSet)
        End Function
        Public Shared Function GetRelatedOneToManyEntities(Service As IOrganizationService, relatedId As Guid, logicalName As String, attributeName As String, columns As ColumnSet) As List(Of Entity)
            Return Service.RetrieveMultiple(logicalName, New ConditionExpression(attributeName, ConditionOperator.Equal, relatedId), columns).Entities.ToList()
        End Function
        Public Shared Function GetRelatedManyToManyEntities(Of T As _BaseProxyClass_)(service As IOrganizationService, proxy As _BaseProxyClass_, linkFromEntityTypeName As String, linkFromAttributeName As String, linkToEntityTypeName As String, linkToAttributeName As String, intermediateEntityName As String, intermediateFromAttributeName As String, intermediateToAttributeName As String, ParamArray columns As String()) As List(Of T)
            Return GetRelatedManyToManyEntities(service, proxy, linkFromEntityTypeName, linkFromAttributeName, linkToEntityTypeName, linkToAttributeName,
                intermediateEntityName, intermediateFromAttributeName, intermediateToAttributeName, columns).ToProxies(Of T)()
        End Function
        Public Shared Function GetRelatedManyToManyEntities(Of T As _BaseProxyClass_)(service As IOrganizationService, proxy As _BaseProxyClass_, linkFromEntityTypeName As String, linkFromAttributeName As String, linkToEntityTypeName As String, linkToAttributeName As String, intermediateEntityName As String, intermediateFromAttributeName As String, intermediateToAttributeName As String, columns As ColumnSet) As List(Of T)
            Return GetRelatedManyToManyEntities(service, proxy, linkFromEntityTypeName, linkFromAttributeName, linkToEntityTypeName, linkToAttributeName,
                intermediateEntityName, intermediateFromAttributeName, intermediateToAttributeName, columns).ToProxies(Of T)()
        End Function
        Public Shared Function GetRelatedManyToManyEntities(service As IOrganizationService, proxy As _BaseProxyClass_, linkFromEntityTypeName As String, linkFromAttributeName As String, linkToEntityTypeName As String, linkToAttributeName As String, intermediateEntityName As String, intermediateFromAttributeName As String, intermediateToAttributeName As String, ParamArray columns As String()) As List(Of Entity)
            Dim columnSet = If((columns.Length = 0), New ColumnSet(True), New ColumnSet(columns))
            Return GetRelatedManyToManyEntities(service, proxy, linkFromEntityTypeName, linkFromAttributeName, linkToEntityTypeName, linkToAttributeName,
                intermediateEntityName, intermediateFromAttributeName, intermediateToAttributeName, columnSet)
        End Function
        Public Shared Function GetRelatedManyToManyEntities(service As IOrganizationService, proxy As _BaseProxyClass_, linkFromEntityTypeName As String, linkFromAttributeName As String, linkToEntityTypeName As String, linkToAttributeName As String,
            intermediateEntityName As String, intermediateFromAttributeName As String, intermediateToAttributeName As String, columns As ColumnSet) As List(Of Entity)
            Dim query = New QueryExpression() With {
                .EntityName = linkToEntityTypeName,
                .ColumnSet = columns
            }

            query.LinkEntities.Add(New LinkEntity() With {
                .LinkFromAttributeName = linkToAttributeName,
                .LinkFromEntityName = linkToEntityTypeName,
                .LinkToAttributeName = intermediateToAttributeName,
                .LinkToEntityName = intermediateEntityName
            })

            query.LinkEntities(0).LinkEntities.Add(New LinkEntity() With {
                .LinkFromAttributeName = intermediateFromAttributeName,
                .LinkFromEntityName = intermediateEntityName,
                .LinkToAttributeName = linkFromAttributeName,
                .LinkToEntityName = linkFromEntityTypeName
            })

            query.LinkEntities(0).LinkEntities(0).LinkCriteria = New FilterExpression()
            query.LinkEntities(0).LinkEntities(0).LinkCriteria.Conditions.Add(New ConditionExpression(linkFromAttributeName, ConditionOperator.Equal, New Object() {proxy.Id}))
            Return service.RetrieveMultipleAllPages(query).Entities.ToList()
        End Function
        Protected Class AttributeEqualityComparer
            Implements IEqualityComparer
            Public Overloads Function Equals(x As Object, y As Object) As Boolean Implements IEqualityComparer.Equals
                If (x Is Nothing OrElse (x.[GetType]() Is GetType(String) AndAlso String.IsNullOrEmpty(TryCast(x, String)))) AndAlso (y Is Nothing OrElse (y.[GetType]() Is GetType(String) AndAlso String.IsNullOrEmpty(TryCast(y, String)))) Then
                    Return True
                Else
                    If x Is Nothing AndAlso y Is Nothing Then
                        Return True
                    ElseIf x Is Nothing AndAlso y IsNot Nothing Then
                        Return False
                    ElseIf x IsNot Nothing AndAlso y Is Nothing Then
                        Return False
                    ElseIf x.GetType() Is y.GetType() Then
                        If x.GetType() Is GetType(OptionSetValue) Then
                            Return DirectCast(x, OptionSetValue).Value = DirectCast(y, OptionSetValue).Value
                        ElseIf x.GetType() Is GetType(BooleanManagedProperty) Then
                            Return DirectCast(x, BooleanManagedProperty).Value = DirectCast(y, BooleanManagedProperty).Value
                        ElseIf x.[GetType]() Is GetType(EntityReference) Then
                            If DirectCast(x, EntityReference).LogicalName = DirectCast(y, EntityReference).LogicalName Then
                                Return DirectCast(x, EntityReference).Id = DirectCast(y, EntityReference).Id
                            Else
                                Return False
                            End If
                        ElseIf x.GetType() Is GetType(Money) Then
                            Return (DirectCast(x, Money).Value = DirectCast(y, Money).Value)
                        ElseIf x.[GetType]() Is GetType(DateTime) OrElse x.[GetType]() Is GetType(System.Nullable(Of DateTime)) Then
                            'Compare only down to the second since CRM only tracks down to the second
                            Return Math.Abs((DirectCast(x, DateTime) - DirectCast(y, DateTime)).TotalSeconds) < 1
                        Else
                            Return x.Equals(y)
                        End If
                    Else
                        Return False
                    End If
                End If
            End Function
            Public Overloads Function GetHashCode(obj As Object) As Integer Implements IEqualityComparer.GetHashCode
                Return obj.GetHashCode()
            End Function
        End Class
        Public Enum eTextOptions
            ''' <summary>Ignore and let CRM handle any issues with the value</summary>
            Ignore
            ''' <summary>If the length is greater than the max length, truncate the value to the max length</summary>
            Truncate
            ''' <summary>Throw an error if the length of the value is greater than the max length</summary>
            ThrowError
        End Enum
        Public Enum eNumberOptions
            ''' <summary>Ignore and let CRM handle any issues with the value.</summary>
            Ignore
            ''' <summary>If the value is less than the min value set the value as the min value.<para>Let CRM handle any issues with the max value.</para></summary>
            CorrectMinIgnoreMax
            ''' <summary>If the value is less than the min value set the value as the min value.<para>Throw an error if the value is greater than the max value.</para></summary>
            CorrectMinThrowMax
            ''' <summary>If the value is greater than the max value set the value as the max value.<para>Let CRM handle any issues with the min value.</para></summary>
            CorrectMaxIgnoreMin
            ''' <summary>If the value is greater than the max value set the value as the max value.<para>Throw an error if the value is less than the min value.</para></summary>
            CorrectMaxThrowMin
            ''' <summary>If the value is less than the min value set the value as the min value.<para>If the value is greater than the max value set the value as the max value.</para></summary>
            CorrectMinAndMax
            ''' <summary>Throw an error if the value is less than the min or greater than the max</summary>
            ThrowError
        End Enum
        Public Enum eErrorType
            Text
            Number
        End Enum
        Public Event PropertyChanged As PropertyChangedEventHandler Implements INotifyPropertyChanged.PropertyChanged
        Public Event PropertyChanging As PropertyChangingEventHandler Implements INotifyPropertyChanging.PropertyChanging
    End Class

    Public Module ExtensionMethods
        <System.Runtime.CompilerServices.Extension>
        Public Function Create(service As IOrganizationService, proxy As _BaseProxyClass_) As Guid
            Dim entityId = service.Create(proxy)
            proxy.Id = entityId
            Return entityId
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Sub Update(service As IOrganizationService, proxy As _BaseProxyClass_)
            proxy.Update(service)
        End Sub
        <System.Runtime.CompilerServices.Extension>
        Public Sub Delete(service As IOrganizationService, proxy As _BaseProxyClass_)
            service.Delete(proxy.LogicalName, proxy.Id)
        End Sub
        <System.Runtime.CompilerServices.Extension>
        Public Sub SetState(service As IOrganizationService, proxy As _BaseProxyClass_, state As Integer, status As Integer)
            service.SetState(proxy, New OptionSetValue(state), New OptionSetValue(status))
        End Sub
        <System.Runtime.CompilerServices.Extension>
        Public Sub SetState(service As IOrganizationService, proxy As _BaseProxyClass_, state As OptionSetValue, status As OptionSetValue)
            Dim request As New SetStateRequest() With {.EntityMoniker = proxy, .State = state, .Status = status}
            service.Execute(request)
        End Sub
        <System.Runtime.CompilerServices.Extension>
        Public Async Function CreateAsync(service As IOrganizationService, entity As Entity) As System.Threading.Tasks.Task(Of Guid)
            Return Await System.Threading.Tasks.Task.Run(Function()
                                                             Dim entityId As Guid = service.Create(entity)
                                                             entity.Id = entityId
                                                             Return entityId
                                                         End Function)
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Async Function UpdateAsync(service As IOrganizationService, entity As Entity) As System.Threading.Tasks.Task
            Await System.Threading.Tasks.Task.Run(Sub()
                                                      service.Update(entity)
                                                  End Sub)
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Async Function DeleteAsync(service As IOrganizationService, entity As Entity) As System.Threading.Tasks.Task
            Await System.Threading.Tasks.Task.Run(Sub()
                                                      service.Delete(entity.LogicalName, entity.Id)
                                                  End Sub)
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Async Function ExecuteAsync(service As IOrganizationService, request As OrganizationRequest) As System.Threading.Tasks.Task(Of OrganizationResponse)
            Return Await System.Threading.Tasks.Task.Run(Function()
                                                             Return service.Execute(request)
                                                         End Function)
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Async Function SetStateAsync(service As IOrganizationService, proxy As _BaseProxyClass_, state As Integer, status As Integer) As System.Threading.Tasks.Task
            Await service.SetStateAsync(proxy, New OptionSetValue(state), New OptionSetValue(status))
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Async Function SetStateAsync(service As IOrganizationService, proxy As _BaseProxyClass_, state As OptionSetValue, status As OptionSetValue) As System.Threading.Tasks.Task
            Dim request = New SetStateRequest() With {
                .EntityMoniker = proxy,
                .State = state,
                .Status = status
            }
            Await service.ExecuteAsync(request)
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function ToProxy(Of T As _BaseProxyClass_)(reference As EntityReference) As T
            If (reference Is Nothing) Then
                Return Nothing
            End If
            Return reference.ToEntity().ToProxy(Of T)()
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function ToEntity(reference As EntityReference) As Entity
            If (reference Is Nothing) Then
                Return Nothing
            End If
            Return New Entity() With {
                .LogicalName = reference.LogicalName,
                .Id = reference.Id
            }
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function ToProxy(Of T As _BaseProxyClass_)(entity As Entity) As T
            If entity IsNot Nothing Then
                Dim constructor As ConstructorInfo = GetType(T).GetConstructor(New Type() {GetType(Entity)})
                Return entity.ToProxy(Of T)(constructor)
            End If
            Return Nothing
        End Function
        <System.Runtime.CompilerServices.Extension>
        Private Function ToProxy(Of T As _BaseProxyClass_)(entity As Entity, construcor As ConstructorInfo) As T
            Return TryCast(construcor.Invoke(New Object() {entity}), T)
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function GetAttributeMetadata(service As IOrganizationService, entityLogicalName As String, attributeLogicalName As String) As AttributeMetadata
            Dim request As New RetrieveAttributeRequest() With {
                .EntityLogicalName = entityLogicalName,
                .LogicalName = attributeLogicalName
            }
            Return TryCast(service.Execute(request), RetrieveAttributeResponse).AttributeMetadata
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function GetOptionMetadata(value As OptionSetValue, service As IOrganizationService, entity As _BaseProxyClass_, attributeLogicalName As String) As OptionMetadata
            Dim attributeMeta As AttributeMetadata = service.GetAttributeMetadata(entity.LogicalName, attributeLogicalName)
            If TypeOf attributeMeta Is EnumAttributeMetadata Then
                Return DirectCast(attributeMeta, EnumAttributeMetadata).GetOptionMetadata(value.Value)
            Else
                Throw New Exception("The attribute is not an Enum type attribute")
            End If
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function GetOptionMetadata(enumMeta As EnumAttributeMetadata, value As Integer) As OptionMetadata
            Return (From m In enumMeta.OptionSet.Options Where m.Value = value Select m).FirstOrDefault()
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function GetOptionSetText(enumMeta As EnumAttributeMetadata, value As Integer) As String
            Return enumMeta.GetOptionMetadata(value).GetOptionSetText()
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function GetOptionSetText(value As OptionSetValue, service As IOrganizationService, entity As _BaseProxyClass_, attributeName As String) As String
            Dim optionMeta As OptionMetadata = GetOptionMetadata(value, service, entity, attributeName)
            Return optionMeta.GetOptionSetText()
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function GetOptionSetText(optionMeta As OptionMetadata) As String
            If optionMeta IsNot Nothing AndAlso optionMeta.Label IsNot Nothing AndAlso optionMeta.Label.UserLocalizedLabel IsNot Nothing Then
                Return optionMeta.Label.UserLocalizedLabel.Label
            End If
            Return String.Empty
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function RetrieveProxies(Of T As _BaseProxyClass_)(service As IOrganizationService, expression As ConditionExpression, columns As ColumnSet) As List(Of T)
            Dim filterExpression As New FilterExpression()
            filterExpression.AddCondition(expression)
            Return service.RetrieveProxies(Of T)(filterExpression, columns)
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function RetrieveProxies(Of T As _BaseProxyClass_)(service As IOrganizationService, expression As FilterExpression, columns As ColumnSet) As List(Of T)
            Dim query As New QueryExpression(_BaseProxyClass_.GetLogicalName(Of T)()) With {
                .ColumnSet = columns,
                .Criteria = expression
            }
            Return service.RetrieveProxies(Of T)(query)
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function RetrieveProxies(Of T As _BaseProxyClass_)(service As IOrganizationService, ParamArray columns As String()) As List(Of T)
            Return service.RetrieveProxies(Of T)(GetColumnSet(columns))
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function RetrieveProxies(Of T As _BaseProxyClass_)(service As IOrganizationService, columns As ColumnSet) As List(Of T)
            Dim query As New QueryExpression(_BaseProxyClass_.GetLogicalName(Of T)()) With {
                .ColumnSet = columns
            }
            Return service.RetrieveProxies(Of T)(query)
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function RetrieveProxies(Of T As _BaseProxyClass_)(service As IOrganizationService, query As QueryExpression) As List(Of T)
            Dim proxies = service.RetrieveMultipleAllPages(query).ToProxies(Of T)
            If (If(query.ColumnSet?.AllColumns, False) = True) Then
                For Each proxy In proxies
                    proxy.AllColumnsRetrievedFromCrm = True
                Next
            End If
            Return proxies
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function ToProxies(Of T As _BaseProxyClass_)(entities As EntityCollection) As List(Of T)
            Return entities.Entities.ToProxies(Of T)()
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function ToProxies(Of T As _BaseProxyClass_)(entities As IEnumerable(Of Entity)) As List(Of T)
            Dim constructor As ConstructorInfo = GetType(T).GetConstructor(New Type() {GetType(Entity)})
            Return (From entity In entities Select entity.ToProxy(Of T)(constructor)).ToList()
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function RetrieveProxy(Of T As _BaseProxyClass_)(entity As EntityReference, service As IOrganizationService, ParamArray columns As String()) As T
            Return service.RetrieveProxy(Of T)(entity.Id, columns)
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function RetrieveProxy(Of T As _BaseProxyClass_)(entity As EntityReference, service As IOrganizationService, columns As ColumnSet) As T
            Return service.RetrieveProxy(Of T)(entity.Id, columns)
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function RetrieveProxy(Of T As _BaseProxyClass_)(service As IOrganizationService, id As Guid, ParamArray columns As String()) As T
            Return service.RetrieveProxy(Of T)(id, GetColumnSet(columns))
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function RetrieveProxy(Of T As _BaseProxyClass_)(service As IOrganizationService, id As Guid, columns As ColumnSet) As T
            Dim entityName As String = _BaseProxyClass_.GetLogicalName(Of T)()
            If (columns Is Nothing) Then
                columns = New ColumnSet(True)
            End If
            Dim proxy = service.Retrieve(entityName, id, columns).ToProxy(Of T)()
            If (columns.AllColumns) Then
                proxy.AllColumnsRetrievedFromCrm = True
            End If
            Return proxy
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function FindProxy(Of T As _BaseProxyClass_)(service As IOrganizationService, propertyName As String, value As Object, ParamArray columns As String()) As T
            Return service.FindProxy(Of T)(propertyName, value, GetColumnSet(columns))
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function FindProxy(Of T As _BaseProxyClass_)(service As IOrganizationService, propertyName As String, value As Object, columns As ColumnSet) As T
            Return service.FindProxies(Of T)(propertyName, value, columns).FirstOrDefault()
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function FindProxy(Of T As _BaseProxyClass_)(service As IOrganizationService, propertyName As String, [operator] As ConditionOperator, value As Object, ParamArray columns As String()) As T
            Return service.FindProxy(Of T)(propertyName, [operator], value, GetColumnSet(columns))
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function FindProxy(Of T As _BaseProxyClass_)(service As IOrganizationService, propertyName As String, [operator] As ConditionOperator, value As Object, columns As ColumnSet) As T
            Return service.FindProxies(Of T)(propertyName, [operator], value, columns).FirstOrDefault()
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function FindProxies(Of T As _BaseProxyClass_)(service As IOrganizationService, propertyName As String, value As Object, ParamArray columns As String()) As List(Of T)
            Return service.FindProxies(Of T)(propertyName, value, GetColumnSet(columns))
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function FindProxies(Of T As _BaseProxyClass_)(service As IOrganizationService, propertyName As String, Value As Object, columns As ColumnSet) As List(Of T)
            Return service.FindProxies(Of T)(propertyName, ConditionOperator.Equal, Value, columns)
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function FindProxies(Of T As _BaseProxyClass_)(service As IOrganizationService, propertyName As String, [operator] As ConditionOperator, value As Object, ParamArray columns As String()) As List(Of T)
            Return service.FindProxies(Of T)(propertyName, [operator], value, GetColumnSet(columns))
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function FindProxies(Of T As _BaseProxyClass_)(service As IOrganizationService, propertyName As String, [operator] As ConditionOperator, value As Object, columns As ColumnSet) As List(Of T)
            Dim propertyCondition As ConditionExpression = If(value IsNot Nothing, New ConditionExpression(propertyName, [operator], value), New ConditionExpression(propertyName, [operator]))
            Return service.RetrieveProxies(Of T)(propertyCondition, columns)
        End Function
        Friend Function GetColumnSet(ParamArray columns As String()) As ColumnSet
            Dim columnSet As New ColumnSet(columns)
            If Not columnSet.Columns.Any() Then
                columnSet = New ColumnSet(True)
            End If
            Return columnSet
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function RetrieveEntity(entity As EntityReference, service As IOrganizationService, ParamArray columns As String()) As Entity
            Return entity.RetrieveEntity(service, GetColumnSet(columns))
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function RetrieveEntity(entity As EntityReference, service As IOrganizationService, columns As ColumnSet) As Entity
            If columns Is Nothing Then
                columns = New ColumnSet(True)
            End If
            Return service.Retrieve(entity.LogicalName, entity.Id, columns)
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function RetrieveMultiple(service As IOrganizationService, logicalName As String, condition As ConditionExpression, ParamArray columns As String()) As EntityCollection
            Return service.RetrieveMultiple(logicalName, condition, GetColumnSet(columns))
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function RetrieveMultiple(service As IOrganizationService, logicalName As String, condition As ConditionExpression, columns As ColumnSet) As EntityCollection
            Dim query As New QueryExpression(logicalName) With {
                .ColumnSet = columns
            }
            query.Criteria.AddCondition(condition)
            Return service.RetrieveMultipleAllPages(query)
        End Function
        Public RetrieveAllPages As Boolean = True
        <System.Runtime.CompilerServices.Extension>
        Public Function RetrieveMultipleAllPages(service As IOrganizationService, query As QueryExpression) As EntityCollection
            If query.PageInfo.PageNumber = 0 AndAlso query.TopCount Is Nothing Then
                query.PageInfo.PageNumber = 1 'No page Set. Set To 1st page.
            End If
            Dim results As EntityCollection = service.RetrieveMultiple(query)
            While results.MoreRecords AndAlso RetrieveAllPages
                Dim currentResults As IList(Of Entity) = results.Entities
                query.PageInfo.PagingCookie = results.PagingCookie
                query.PageInfo.PageNumber += 1
                results = service.RetrieveMultiple(query)
                results.Entities.AddRange(currentResults)
            End While
            Return results
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Sub RelateEntities(service As IOrganizationService, target As EntityReference, intermediateEntityName As String, Entities As IList(Of Entity))
            service.RelateEntities(target, intermediateEntityName, Nothing, Entities.ToArray())
        End Sub
        <System.Runtime.CompilerServices.Extension>
        Public Sub RelateEntities(service As IOrganizationService, target As EntityReference, intermediateEntityName As String, entityRole As EntityRole?, Entities As IList(Of Entity))
            service.RelateEntities(target, intermediateEntityName, entityRole, Entities.ToArray())
        End Sub
        <System.Runtime.CompilerServices.Extension>
        Public Sub RelateEntities(service As IOrganizationService, target As EntityReference, intermediateEntityName As String, entityRole As EntityRole?, ParamArray entities As Entity())
            Dim references As IList(Of EntityReference) = (From entity In entities Select entity.ToEntityReference()).ToList()
            service.RelateEntities(target, intermediateEntityName, entityRole, references)
        End Sub
        <System.Runtime.CompilerServices.Extension>
        Public Sub RelateEntities(service As IOrganizationService, target As EntityReference, intermediateEntityName As String, entityRole As EntityRole?, entities As IList(Of EntityReference))
            service.RelateEntities(target, intermediateEntityName, entityRole, entities.ToArray())
        End Sub
        <System.Runtime.CompilerServices.Extension>
        Public Sub RelateEntities(service As IOrganizationService, target As EntityReference, intermediateEntityName As String, ParamArray entities As EntityReference())
            service.RelateEntities(target, intermediateEntityName, Nothing, entities)
        End Sub
        <System.Runtime.CompilerServices.Extension>
        Public Sub RelateEntities(service As IOrganizationService, target As EntityReference, intermediateEntityName As String, entityRole As EntityRole?, ParamArray entities As EntityReference())
            Dim request As New AssociateRequest() With {
                .Relationship = New Relationship(intermediateEntityName) With {.PrimaryEntityRole = entityRole},
                .Target = target
            }
            request.RelatedEntities = New EntityReferenceCollection(entities)
            service.Execute(request)
        End Sub
        <System.Runtime.CompilerServices.Extension>
        Public Sub UnRelateEntities(service As IOrganizationService, target As EntityReference, intermediateEntityName As String, entities As IList(Of Entity))
            service.UnRelateEntities(target, intermediateEntityName, Nothing, entities.ToArray())
        End Sub
        <System.Runtime.CompilerServices.Extension>
        Public Sub UnRelateEntities(service As IOrganizationService, target As EntityReference, intermediateEntityName As String, entityRole As EntityRole?, entities As IList(Of Entity))
            service.UnRelateEntities(target, intermediateEntityName, entityRole, entities.ToArray())
        End Sub
        <System.Runtime.CompilerServices.Extension>
        Public Sub UnRelateEntities(service As IOrganizationService, target As EntityReference, intermediateEntityName As String, entityRole As EntityRole?, ParamArray entities As Entity())
            Dim references As IList(Of EntityReference) = (From entity In entities Select entity.ToEntityReference()).ToList()
            service.UnRelateEntities(target, intermediateEntityName, entityRole, references)
        End Sub
        <System.Runtime.CompilerServices.Extension>
        Public Sub UnRelateEntities(service As IOrganizationService, target As EntityReference, intermediateEntityName As String, entityRole As EntityRole?, entities As IList(Of EntityReference))
            service.UnRelateEntities(target, intermediateEntityName, entityRole, entities.ToArray())
        End Sub
        <System.Runtime.CompilerServices.Extension>
        Public Sub UnRelateEntities(service As IOrganizationService, target As EntityReference, intermediateEntityName As String, ParamArray entities As EntityReference())
            service.UnRelateEntities(target, intermediateEntityName, Nothing, entities)
        End Sub
        <System.Runtime.CompilerServices.Extension>
        Public Sub UnRelateEntities(service As IOrganizationService, target As EntityReference, intermediateEntityName As String, entityRole As EntityRole?, ParamArray entities As EntityReference())
            Dim request As New DisassociateRequest() With {
                .Target = target,
                .Relationship = New Relationship(intermediateEntityName) With {.PrimaryEntityRole = entityRole}
            }
            request.RelatedEntities = New EntityReferenceCollection(entities)
            service.Execute(request)
        End Sub
        <System.Runtime.CompilerServices.Extension>
        Public Sub Add(requests As List(Of OrganizationRequest), proxy As _BaseProxyClass_)
            Dim request As OrganizationRequest = GetOrganizationRequest(proxy)
            If request IsNot Nothing Then
                requests.Add(request)
            End If
        End Sub
        <System.Runtime.CompilerServices.Extension>
        Public Sub Add(collection As OrganizationRequestCollection, proxy As _BaseProxyClass_)
            Dim reqeust As OrganizationRequest = GetOrganizationRequest(proxy)
            If reqeust IsNot Nothing Then
                collection.Add(reqeust)
            End If
        End Sub
        Public Function GetOrganizationRequest(proxy As _BaseProxyClass_) As OrganizationRequest
            If proxy.Id <> Guid.Empty AndAlso proxy.IsDirty Then
                Return New UpdateRequest() With {
                    .Target = proxy.GetChangedEntity()
                }
            ElseIf proxy.Id = Guid.Empty Then
                Return New CreateRequest() With {
                    .Target = proxy
                }
            End If
            Return Nothing
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Sub GrantAccess(entity As IAccessControlledEntity, service As IOrganizationService, principal As EntityReference, accessMask As AccessRights)
            Dim request As New GrantAccessRequest() With {
                .PrincipalAccess = New PrincipalAccess() With {
                    .Principal = principal,
                    .AccessMask = accessMask
                },
                .Target = DirectCast(DirectCast(entity, Object), EntityReference)
            }
            service.Execute(request)
        End Sub
        <System.Runtime.CompilerServices.Extension>
        Public Sub ModifyAccess(entity As IAccessControlledEntity, service As IOrganizationService, principal As EntityReference, accessMask As AccessRights)
            Dim request As New ModifyAccessRequest() With {
                .PrincipalAccess = New PrincipalAccess() With {
                    .Principal = principal,
                    .AccessMask = accessMask
                },
                .Target = DirectCast(DirectCast(entity, Object), EntityReference)
            }
            service.Execute(request)
        End Sub
        <System.Runtime.CompilerServices.Extension>
        Public Sub RevokeAccess(entity As IAccessControlledEntity, service As IOrganizationService, revokee As EntityReference)
            Dim request As New RevokeAccessRequest() With {
                .Revokee = revokee,
                .Target = DirectCast(DirectCast(entity, Object), EntityReference)
            }
            service.Execute(request)
        End Sub
        Private _cachedEntities As New Dictionary(Of String, Dictionary(Of Guid, Entity))()
        Public Sub ClearCache(Optional entityLogicalName As String = Nothing)
            SyncLock _cachedEntities
                If Not String.IsNullOrEmpty(entityLogicalName) Then
                    If _cachedEntities.ContainsKey(entityLogicalName) Then
                        _cachedEntities(entityLogicalName) = New Dictionary(Of Guid, Entity)()
                    End If
                Else
                    _cachedEntities.Clear()
                End If
            End SyncLock
        End Sub
        <System.Runtime.CompilerServices.Extension>
        Public Function RetrieveCachedEntityById(service As IOrganizationService, EntieyLogicalName As String, id As Guid, columns As ColumnSet) As Entity
            SyncLock _cachedEntities
                Dim cachedEntity As Entity = GetEntityFromCache(EntieyLogicalName, id)
                If cachedEntity Is Nothing Then
                    cachedEntity = service.Retrieve(EntieyLogicalName, id, columns)
                    AddEntityToCache(cachedEntity)
                End If
                Return cachedEntity
            End SyncLock
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function GetEntityFromCache(service As IOrganizationService, entityLogicalName As String, Id As Guid) As Entity
            Return GetEntityFromCache(entityLogicalName, Id)
        End Function
        Public Function GetEntityFromCache(EntityLogicalName As String, Id As Guid) As Entity
            SyncLock _cachedEntities
                Dim cachedEntity As Entity = Nothing
                Dim cachedEntities As Dictionary(Of Guid, Entity) = GetEntitiesFromCache(EntityLogicalName)
                cachedEntities.TryGetValue(Id, cachedEntity)
                Return cachedEntity
            End SyncLock
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function GetEntitiesFromCache(service As IOrganizationService, entityLogicalName As String) As Dictionary(Of Guid, Entity)
            Return GetEntitiesFromCache(entityLogicalName)
        End Function
        Public Function GetEntitiesFromCache(entityLogicalName As String) As Dictionary(Of Guid, Entity)
            SyncLock _cachedEntities
                Dim cachedEntities As New Dictionary(Of Guid, Entity)()
                _cachedEntities.TryGetValue(entityLogicalName, cachedEntities)
                Return If(cachedEntities IsNot Nothing, cachedEntities, New Dictionary(Of Guid, Entity)())
            End SyncLock
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Sub AddEntitiesToCache(Of T As _BaseProxyClass_)(service As IOrganizationService, entities As ICollection(Of T))
            For Each item As T In entities
                AddEntityToCache(item)
            Next
        End Sub
        <System.Runtime.CompilerServices.Extension>
        Public Sub AddEntitiesToCache(service As IOrganizationService, entities As ICollection(Of Entity))
            AddEntitiesToCache(entities)
        End Sub
        Public Sub AddEntitiesToCache(entities As ICollection(Of Entity))
            SyncLock _cachedEntities
                For Each item As Entity In entities
                    AddEntityToCache(item)
                Next
            End SyncLock
        End Sub
        <System.Runtime.CompilerServices.Extension>
        Public Sub AddEntityToCache(service As IOrganizationService, item As Entity)
            AddEntityToCache(item)
        End Sub
        Public Sub AddEntityToCache(item As Entity)
            SyncLock _cachedEntities
                If Not _cachedEntities.ContainsKey(item.LogicalName) Then
                    _cachedEntities.Add(item.LogicalName, New Dictionary(Of Guid, Entity)() From {
                    })
                End If
                If Not _cachedEntities(item.LogicalName).ContainsKey(item.Id) Then
                    _cachedEntities(item.LogicalName).Add(item.Id, item)
                End If
            End SyncLock
        End Sub
        <System.Runtime.CompilerServices.Extension>
        Public Function RetrieveCachedProxy(Of T As _BaseProxyClass_)(service As IOrganizationService, entity As EntityReference) As T
            Return RetrieveCachedProxy(Of T)(service, entity, New ColumnSet(True))
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function RetrieveCachedProxy(Of T As _BaseProxyClass_)(service As IOrganizationService, entity As EntityReference, ParamArray columns As String()) As T
            Return RetrieveCachedProxy(Of T)(service, entity, GetColumnSet(columns))
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function RetrieveCachedProxy(Of T As _BaseProxyClass_)(service As IOrganizationService, entity As EntityReference, columns As ColumnSet) As T
            Return RetrieveCachedEntityById(service, entity.LogicalName, entity.Id, columns).ToProxy(Of T)()
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function RetrieveCachedProxy(Of T As _BaseProxyClass_)(entity As EntityReference, service As IOrganizationService) As T
            Return RetrieveCachedProxy(Of T)(entity, service, New ColumnSet(True))
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function RetrieveCachedProxy(Of T As _BaseProxyClass_)(entity As EntityReference, service As IOrganizationService, ParamArray columns As String()) As T
            Return RetrieveCachedProxy(Of T)(entity, service, GetColumnSet(columns))
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function RetrieveCachedProxy(Of T As _BaseProxyClass_)(entity As EntityReference, service As IOrganizationService, columns As ColumnSet) As T
            Return service.RetrieveCachedProxy(Of T)(entity, columns)
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Async Function GetEntityMetadataAsync(service As IOrganizationService, logicalName As String, filters As EntityFilters) As System.Threading.Tasks.Task(Of EntityMetadata)
            Return Await System.Threading.Tasks.Task.Run(Function()
                                                             Return GetEntityMetadata(service, logicalName, filters)
                                                         End Function)
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Function GetEntityMetadata(service As IOrganizationService, logicalName As String, filters As EntityFilters) As EntityMetadata
            Dim request As New RetrieveEntityRequest() With {
                .LogicalName = logicalName,
                .EntityFilters = filters
            }
            Return DirectCast(service.Execute(request), RetrieveEntityResponse).EntityMetadata
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Async Function RetrieveAsync(service As IOrganizationService, entityName As String, id As Guid, ParamArray columns As String()) As System.Threading.Tasks.Task(Of Entity)
            Return Await service.RetrieveAsync(entityName, id, GetColumnSet(columns))
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Async Function RetrieveAsync(service As IOrganizationService, entity As EntityReference, ParamArray columns As String()) As System.Threading.Tasks.Task(Of Entity)
            Return Await service.RetrieveAsync(entity.LogicalName, entity.Id, GetColumnSet(columns))
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Async Function RetrieveAsync(service As IOrganizationService, entity As EntityReference, columns As ColumnSet) As System.Threading.Tasks.Task(Of Entity)
            Return Await service.RetrieveAsync(entity.LogicalName, entity.Id, columns)
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Async Function RetrieveAsync(service As IOrganizationService, entityName As String, id As Guid, columns As ColumnSet) As System.Threading.Tasks.Task(Of Entity)
            Return Await System.Threading.Tasks.Task.Run(Function()
                                                             Return service.Retrieve(entityName, id, columns)
                                                         End Function)
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Sub CopyTo(entity As Entity, toEntity As Entity, Optional overwriteMissingValuesOnly As Boolean = True)
            If entity.Id <> Guid.Empty AndAlso (toEntity.Id = Guid.Empty OrElse Not overwriteMissingValuesOnly) Then
                toEntity.Id = entity.Id
            End If
            For Each attribute As KeyValuePair(Of String, Object) In entity.Attributes
                If Not toEntity.Contains(attribute.Key) OrElse Not overwriteMissingValuesOnly Then
                    toEntity(attribute.Key) = attribute.Value
                End If
            Next
        End Sub
        <System.Runtime.CompilerServices.Extension>
        Public Async Function RetrieveMultipleAsync(service As IOrganizationService, query As QueryExpression) As System.Threading.Tasks.Task(Of EntityCollection)
            Return Await System.Threading.Tasks.Task.Run(Function()
                                                             Return service.RetrieveMultiple(query)
                                                         End Function)
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Async Function RetrieveMultipleAsync(service As IOrganizationService, entityName As String, condition As ConditionExpression, columns As ColumnSet) As System.Threading.Tasks.Task(Of EntityCollection)
            Return Await System.Threading.Tasks.Task.Run(Function()
                                                             Return service.RetrieveMultiple(entityName, condition, columns)
                                                         End Function)
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Async Function RetrieveMultipleAsync(service As IOrganizationService, entityName As String, condition As ConditionExpression, ParamArray columns As String()) As System.Threading.Tasks.Task(Of EntityCollection)
            Return Await System.Threading.Tasks.Task.Run(Function()
                                                             Return service.RetrieveMultiple(entityName, condition, columns)
                                                         End Function)
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Async Function DeleteAsync(service As IOrganizationService, entity As EntityReference) As System.Threading.Tasks.Task
            Await DeleteAsync(service, entity.LogicalName, entity.Id)
        End Function
        <System.Runtime.CompilerServices.Extension>
        Public Async Function DeleteAsync(service As IOrganizationService, logicalName As String, id As Guid) As System.Threading.Tasks.Task
            Await System.Threading.Tasks.Task.Run(Sub()
                                                      service.Delete(logicalName, id)
                                                  End Sub)
        End Function
    End Module
    Public Interface IAccessControlledEntity
    End Interface
End Namespace